from __future__ import annotations

from scoda.enumerations.message_type import MessageType
from scoda.misc.music_theory import Key


class Message:
    """Class representing a musical message.
    """

    def __init__(self,
                 message_type: MessageType = None,
                 channel: int = None,
                 time: int = None,
                 note: int = None,
                 velocity: int = None,
                 control: int = None,
                 numerator: int = None,
                 denominator: int = None,
                 key: Key = None,
                 program: int = None) -> None:
        super().__init__()
        self.message_type = message_type
        self.channel = channel
        self.time = time
        self.note = note
        self.velocity = velocity
        self.control = control
        self.program = program
        self.numerator = numerator
        self.denominator = denominator
        self.key = key

        # Defaults

        if self.channel is None:
            self.channel = 0

    def equivalent(self, other) -> bool:
        if not isinstance(other, Message):
            return False

        if self.__dict__.keys() != other.__dict__.keys():
            return False

        for self_field, other_field in zip(list(self.__dict__.values()), list(other.__dict__.values())):
            if not self_field == other_field:
                return False

        return True

    def copy(self) -> Message:
        cpy = self.__class__(
            message_type=self.message_type,
            channel=self.channel,
            time=self.time,
            note=self.note,
            velocity=self.velocity,
            control=self.control,
            program=self.program,
            numerator=self.numerator,
            denominator=self.denominator,
            key=self.key
        )
        return cpy

    def __repr__(self) -> str:
        representation = f"({self.message_type.value}:"

        attributes = [
            ("time", self.time),
            ("channel", self.channel),
            ("note", self.note),
            ("velocity", self.velocity),
            ("control", self.control),
            ("program", self.program),
            ("numerator", self.numerator),
            ("denominator", self.denominator),
            ("key", self.key.value if self.key is not None else None)
        ]

        for attr_name, attr_value in attributes:
            if attr_value is not None:
                representation += f", {attr_name}={attr_value}"

        representation += ")"
        representation = representation.replace(",", "", 1)

        return representation

    @staticmethod
    def from_dict(dictionary: dict) -> Message:
        message_type = dictionary.get("message_type", None)
        if message_type is not None:
            message_type = MessageType[message_type]

        key = dictionary.get("key", None)
        if key is not None:
            key = Key(key)

        msg = Message(message_type=message_type,
                      channel=dictionary.get("channel", None),
                      note=dictionary.get("note", None),
                      velocity=dictionary.get("velocity", None),
                      control=dictionary.get("control", None),
                      program=dictionary.get("program", None),
                      numerator=dictionary.get("numerator", None),
                      denominator=dictionary.get("denominator", None),
                      key=key,
                      time=dictionary.get("time", None))

        return msg


class ReadOnlyMessage(Message):

    def __init__(self, message: Message):
        super().__init__(
            message_type=message.message_type,
            channel=message.channel,
            time=message.time,
            note=message.note,
            velocity=message.velocity,
            control=message.control,
            program=message.program,
            numerator=message.numerator,
            denominator=message.denominator,
            key=message.key
        )
        self.__dict__["_initialised"] = True

    @property
    def initialised(self):
        return self.__dict__.get("_initialised", False)

    def __setattr__(self, attribute, value):
        if self.initialised:
            raise AttributeError("This object is read-only")
        else:
            super().__setattr__(attribute, value)

    def __delattr__(self, attribute):
        raise AttributeError("This object is read-only")

    def copy(self) -> Message:
        return Message(
            message_type=self.message_type,
            channel=self.channel,
            time=self.time,
            note=self.note,
            velocity=self.velocity,
            control=self.control,
            program=self.program,
            numerator=self.numerator,
            denominator=self.denominator,
            key=self.key
        )

    def __repr__(self):
        repr_super = super().__repr__()
        return repr_super[:1] + "read only: " + repr_super[1:]



from __future__ import annotations

from abc import ABC, abstractmethod

from scoda.elements.message import Message


class AbstractSequence(ABC):
    """Class representing an abstract musical sequence.

    """

    def __init__(self, messages: list = None) -> None:
        super().__init__()
        self._messages = []

        if messages is not None:
            self._messages.extend(messages)

    def copy(self) -> AbstractSequence:
        cpy = self.__class__(messages=[msg.copy() for msg in self._messages])
        return cpy

    @abstractmethod
    def add_message(self, msg: Message) -> None:
        """Adds a message to the sequence.

        For an `AbsoluteSequence` the message has to contain an entry for `time`, according to which it will
        be sorted into the sequence. This is not needed for a `RelativeSequence`.

        Args:
            msg: The message to append

        """
        pass



import json
from pathlib import Path

# General settings

"""Parts per quarter note, amount of ticks used for the representation of a quarter note"""
PPQN: int
"""Maximum velocity value, velocity values usually range from 0 to 127"""
VELOCITY_MAX: int
"""Amount of velocity bins to use, consolidating velocity values into `VELOCITY_BINS` bins"""
VELOCITY_BINS: int
"""MIDI value of lowest note"""
NOTE_LOWER_BOUND: int
"""MIDI value of highest note"""
NOTE_UPPER_BOUND: int
"""Maximum power of 2 defining lower length of notes, e.g., 8 corresponds to PPQN / 2^3, or thirty-second notes"""
NOTE_VALUE_LOWER_BOUND: int
"""Maximum power of 2 defining upper length of notes, e.g., 8 corresponds to PPQN * 2^3, or double whole notes"""
NOTE_VALUE_UPPER_BOUND: int
"""Amount of dots for dotted notes to consider"""
DOTTED_ITERATIONS: int
"""Considered tuplets, e.g., a value of (3, 2) allows for triplets"""
VALID_TUPLETS: list[tuple[int, int]]
"""The default time signature numerator"""
DEFAULT_TIME_SIGNATURE_NUMERATOR: int
"""The default time signature denominator"""
DEFAULT_TIME_SIGNATURE_DENOMINATOR: int

# Pattern Recognition

"""Minimum length of patterns to be considered"""
PATTERN_LENGTH_MIN: int
"""Amount of seconds after which greedy pattern matching is applied"""
PATTERN_SECONDS_SEARCH_DURATION: int
"""Regex used to find patterns"""
REGEX_PATTERN: str
"""Regex used to find subpatterns"""
REGEX_SUBPATTERN: str


def load_from_file(path_settings: Path = None):
    if path_settings is None:
        path_settings = Path(__file__).parent.parent.joinpath("config/default_settings.json")

    with open(path_settings) as settings_file:
        settings = json.load(settings_file)

    global PPQN
    global VELOCITY_MAX
    global VELOCITY_BINS
    global NOTE_LOWER_BOUND
    global NOTE_UPPER_BOUND
    global NOTE_VALUE_LOWER_BOUND
    global NOTE_VALUE_UPPER_BOUND
    global DOTTED_ITERATIONS
    global VALID_TUPLETS
    global DEFAULT_TIME_SIGNATURE_NUMERATOR
    global DEFAULT_TIME_SIGNATURE_DENOMINATOR

    global PATTERN_LENGTH_MIN
    global PATTERN_SECONDS_SEARCH_DURATION
    global REGEX_PATTERN
    global REGEX_SUBPATTERN

    PPQN = settings["general_settings"]["ppqn"]
    VELOCITY_MAX = settings["general_settings"]["velocity_max"]
    VELOCITY_BINS = settings["general_settings"]["velocity_bins"]
    NOTE_LOWER_BOUND = settings["general_settings"]["note_lower_bound"]
    NOTE_UPPER_BOUND = settings["general_settings"]["note_upper_bound"]
    NOTE_VALUE_LOWER_BOUND = settings["general_settings"]["note_value_lower_bound"]
    NOTE_VALUE_UPPER_BOUND = settings["general_settings"]["note_value_upper_bound"]
    DOTTED_ITERATIONS = settings["general_settings"]["dotted_iterations"]

    VALID_TUPLETS = []
    for tuplet in settings["general_settings"]["tuplets_valid"]:
        VALID_TUPLETS.append(tuple(tuplet))

    DEFAULT_TIME_SIGNATURE_NUMERATOR = settings["general_settings"]["default_time_signature_numerator"]
    DEFAULT_TIME_SIGNATURE_DENOMINATOR = settings["general_settings"]["default_time_signature_denominator"]

    PATTERN_LENGTH_MIN = settings["pattern_recognition"]["pattern_length_min"]
    PATTERN_SECONDS_SEARCH_DURATION = settings["pattern_recognition"]["pattern_seconds_search_duration"]
    REGEX_PATTERN = settings["pattern_recognition"]["regex_pattern"]
    REGEX_SUBPATTERN = settings["pattern_recognition"]["regex_subpattern"]


load_from_file()



import itertools
import math

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.enumerations.tokenisation_prefixes import TokenisationPrefixes
from scoda.exceptions.tokenisation_exception import TokenisationException
from scoda.misc.music_theory import CircleOfFifths
from scoda.misc.scoda_logging import get_logger
from scoda.misc.util import get_default_step_sizes, get_default_note_values, get_velocity_bins, bin_velocity
from scoda.sequences.sequence import Sequence
from scoda.settings.settings import PPQN, DEFAULT_TIME_SIGNATURE_NUMERATOR, DEFAULT_TIME_SIGNATURE_DENOMINATOR

LOGGER = get_logger(__name__)


class MultiTrackLargeVocabularyNotelikeTokeniser:
    sort_order = [TokenisationPrefixes.TRACK.value, TokenisationPrefixes.VALUE.value,
                  TokenisationPrefixes.VELOCITY.value, TokenisationPrefixes.PITCH.value]

    def __init__(self,
                 ppqn: int = None,
                 num_tracks: int = 1,
                 pitch_range: tuple[int, int] = (21, 108),
                 step_sizes: list[int] = None,
                 note_values: list[int] = None,
                 velocity_bins: int = 1,
                 time_signature_range: tuple[int, int] = (2, 16),
                 bar_position_quarters_range: float = 4.0,
                 flag_absolute_bar_position: bool = False,
                 flag_running_values: bool = True,
                 flag_running_time_signature: bool = True,
                 flag_bar_token: bool = True,
                 flag_fuse_track: bool = True,
                 flag_fuse_value: bool = True,
                 flag_fuse_velocity: bool = True,
                 flag_ignore_time_signature: bool = False,
                 flag_simplify_time_signature: bool = True):
        """A tokeniser following the principles of the large-vocabulary notelike tokeniser with support for multiple tracks.

        Args:
            ppqn (int, optional): Parts per quarter note (default ``None``).
            num_tracks (int, optional): Number of tracks (default ``1``).
            pitch_range (Tuple[int, int], optional): Range of MIDI note numbers (default ``(21, 108)``).
            step_sizes (list[int], optional): List of quantised step sizes (default ``None``).
            note_values (list[int], optional): List of quantised note values (default ``None``).
            velocity_bins (int, optional): Number of velocity bins to quantise velocity to (default ``1``).
            time_signature_range (Tuple[int, int], optional): Range of time signatures (default ``(2, 16)``).
            bar_position_quarters_range (float, optional): Range of maximum supported bar position in quarters (default ``4.0``).
            flag_absolute_bar_position (bool, optional): Flag to indicate if absolute bar position should be used (default ``False``).
            flag_running_values (bool, optional): Flag to indicate if running values should be used (default ``True``).
            flag_running_time_signature (bool, optional): Flag to indicate if running time signature should be used (default ``True``).
            flag_bar_token (bool, optional): Flag to indicate if a token for bar borders should be used (default ``True``).
            flag_fuse_track (bool, optional): Flag to indicate if the track should be fused into the token (default ``True``).
            flag_fuse_value (bool, optional): Flag to indicate if the note value should be fused into the token (default ``True``).
            flag_fuse_velocity (bool, optional): Flag to indicate if the velocity should be fused into the token (default ``True``).
            flag_ignore_time_signature (bool, optional): Flag to indicate if time signature should be ignored (default ``False``).
            flag_simplify_time_signature (bool, optional): Flag to indicate if time signature should be simplified to quarters (default ``True``).

        """
        self.dictionary = dict()
        self.inverse_dictionary = dict()
        self._dictionary_size = 0

        self.ppqn = ppqn
        self.step_sizes = step_sizes
        self.note_values = note_values
        self.num_tracks = num_tracks
        self.pitch_range = pitch_range
        self.time_signature_range = time_signature_range
        self.bar_position_quarters_range = bar_position_quarters_range

        self.flag_absolute_bar_position = flag_absolute_bar_position
        self.flag_running_values = flag_running_values
        self.flag_running_time_signature = flag_running_time_signature
        self.flag_bar_token = flag_bar_token
        self.flag_fuse_track = flag_fuse_track
        self.flag_fuse_value = flag_fuse_value
        self.flag_fuse_velocity = flag_fuse_velocity
        self.flag_ignore_time_signature = flag_ignore_time_signature
        self.flag_simplify_time_signature = flag_simplify_time_signature

        # Default Values
        if self.ppqn is None:
            self.ppqn = PPQN
        if self.step_sizes is None:
            self.step_sizes = get_default_step_sizes(lower_bound_shift=1)
        self.step_sizes.sort()
        if self.note_values is None:
            self.note_values = get_default_note_values()
        self.note_values.sort()

        self.velocity_bins = get_velocity_bins(velocity_bins=velocity_bins)

        # Memory
        self.cur_time = None
        self.cur_rest_buffer = None

        # Construct dictionary
        self._construct_dictionary()

        if not self.flag_running_time_signature:
            raise NotImplementedError()

        if self.flag_absolute_bar_position and not self.flag_bar_token:
            raise TokenisationException("Absolute bar position requires bar token")

    @property
    def dictionary_size(self):
        return self._dictionary_size

    def tokenise(self,
                 sequences_bar: list[Sequence],
                 state_dict: dict = None) -> list[str]:
        if state_dict is None:
            state_dict = dict()

        # Setup Values
        tokens = []
        cur_time = state_dict.get("cur_time", 0)
        cur_time_bar = state_dict.get("cur_time_bar", 0)
        cur_time_signature_numerator = state_dict.get("cur_time_signature_numerator",
                                                      DEFAULT_TIME_SIGNATURE_NUMERATOR)
        cur_time_signature_denominator = state_dict.get("cur_time_signature_denominator",
                                                        DEFAULT_TIME_SIGNATURE_DENOMINATOR)
        cur_bar_capacity_total = int(self.ppqn * 4 * cur_time_signature_numerator / cur_time_signature_denominator)
        cur_bar_capacity_remaining = state_dict.get("cur_bar_capacity_remaining", cur_bar_capacity_total)
        prv_track = state_dict.get("prv_track", -1)
        prv_value = state_dict.get("prv_value", -1)
        prv_velocity = state_dict.get("prv_velocity", -1)
        prv_shift = state_dict.get("cur_time", 0)

        # Sanity check
        if not len(sequences_bar) == self.num_tracks:
            raise TokenisationException("Number of sequences does not match number of tracks")

        # Utility function
        def _apply_rest(rest: int):
            nonlocal tokens
            nonlocal cur_time
            nonlocal cur_time_bar
            nonlocal cur_bar_capacity_remaining
            buf_rest = rest

            # While rest buffer not empty
            while buf_rest > 0:
                if not self.flag_absolute_bar_position:
                    nxt_rest = min(buf_rest, cur_bar_capacity_remaining)

                    # Check if next rest value is valid
                    if not (nxt_rest > self.step_sizes[-1] or any(
                            nxt_rest >= step_size for step_size in self.step_sizes)):
                        raise TokenisationException(f"Invalid remaining rest value: {nxt_rest}")

                    # Check if next rest value is larger than the largest step size
                    if nxt_rest > self.step_sizes[-1]:
                        rest_value = self.step_sizes[-1]
                    else:
                        rest_value = next(step_size for step_size in reversed(self.step_sizes) if nxt_rest >= step_size)

                    # Apply rest
                    cur_time += rest_value
                    cur_time_bar += rest_value
                    cur_bar_capacity_remaining -= rest_value
                    buf_rest -= rest_value
                    tokens.append(f"{TokenisationPrefixes.REST.value}_{rest_value:02}")

                    # Check if we are at bar end
                    if cur_bar_capacity_remaining == 0:
                        if self.flag_bar_token:
                            tokens.append(TokenisationPrefixes.BAR.value)
                        cur_time_bar = 0
                        cur_bar_capacity_remaining = cur_bar_capacity_total
                else:
                    if buf_rest >= cur_bar_capacity_remaining:
                        cur_time += cur_bar_capacity_remaining
                        cur_time_bar = 0
                        buf_rest -= cur_bar_capacity_remaining
                        cur_bar_capacity_remaining = cur_bar_capacity_total
                        tokens.append(TokenisationPrefixes.BAR.value)
                    else:
                        if cur_time_bar + buf_rest > self.bar_position_quarters_range * self.ppqn:
                            raise TokenisationException(f"Invalid bar position value: {cur_time_bar + buf_rest}")

                        cur_time += buf_rest
                        cur_time_bar += buf_rest
                        cur_bar_capacity_remaining -= buf_rest
                        buf_rest = 0
                        tokens.append(f"{TokenisationPrefixes.POSITION.value}_{cur_time_bar:03}")

        # Merge sequences
        for i, sequence_bar in enumerate(sequences_bar):
            sequence_bar.set_channel(i)
        sequence_bar = Sequence()
        sequence_bar.merge(sequences_bar)

        # Get interleaved pairings
        interleaved_pairings = sequence_bar.get_interleaved_message_pairings(
            [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE, MessageType.INTERNAL])

        # Handle pairings
        for interleaved_pairing in interleaved_pairings:
            event_pairing = interleaved_pairing[1]

            msg_channel = interleaved_pairing[0]
            if not msg_channel == event_pairing[0].channel:
                raise TokenisationException("Channel mismatch")

            msg_type = event_pairing[0].message_type
            msg_time = event_pairing[0].time + prv_shift

            # Check if message occurs at current time, if not place rest messages
            if not cur_time == msg_time:
                _apply_rest(msg_time - cur_time)

            # Handle notes
            if msg_type == MessageType.NOTE_ON:
                msg_channel = event_pairing[0].channel
                msg_note = event_pairing[0].note
                msg_value = event_pairing[1].time - event_pairing[0].time
                msg_velocity = self.velocity_bins[bin_velocity(event_pairing[0].velocity, self.velocity_bins)]

                if not (self.pitch_range[0] <= msg_note <= self.pitch_range[1]):
                    raise TokenisationException(f"Invalid note pitch: {msg_note}")
                if msg_value not in self.note_values:
                    raise TokenisationException(f"Invalid note value: {msg_value}")

                token = ""

                if not self.flag_fuse_track and (msg_channel != prv_track or not self.flag_running_values):
                    tokens.append(f"{TokenisationPrefixes.TRACK.value}_{msg_channel:02}")
                elif self.flag_fuse_track:
                    token += f"{TokenisationPrefixes.TRACK.value}_{msg_channel:02}-"

                token += f"{TokenisationPrefixes.PITCH.value}_{msg_note:03}-"

                if not self.flag_fuse_value and (msg_value != prv_value or not self.flag_running_values):
                    tokens.append(f"{TokenisationPrefixes.VALUE.value}_{msg_value:02}")
                elif self.flag_fuse_value:
                    token += f"{TokenisationPrefixes.VALUE.value}_{msg_value:02}-"

                if not self.flag_fuse_velocity and (msg_velocity != prv_velocity or not self.flag_running_values):
                    tokens.append(f"{TokenisationPrefixes.VELOCITY.value}_{msg_velocity:03}")
                elif self.flag_fuse_velocity:
                    token += f"{TokenisationPrefixes.VELOCITY.value}_{msg_velocity:03}-"

                token = token[:-1]
                tokens.append(token)

                prv_track = msg_channel
                prv_value = msg_value
                prv_velocity = msg_velocity
            # Handle time signatures
            elif msg_type == MessageType.TIME_SIGNATURE and not self.flag_ignore_time_signature:
                if cur_time_bar > 0:
                    LOGGER.info(
                        f"Skipping time signature change mid-bar at time {cur_time} (bar time {cur_time_bar})")
                    continue

                msg_numerator = event_pairing[0].numerator
                msg_denominator = event_pairing[0].denominator

                scaled = msg_numerator * (DEFAULT_TIME_SIGNATURE_DENOMINATOR / msg_denominator)
                if not float(scaled).is_integer():
                    raise TokenisationException(
                        f"Time signature {int(msg_numerator)}/{int(msg_denominator)} cannot be represented as multiples of eights")
                scaled = int(scaled)
                if not self.time_signature_range[0] <= scaled <= self.time_signature_range[1]:
                    raise TokenisationException(f"Invalid time signature numerator: {scaled}")

                cur_time_signature_numerator = msg_numerator
                cur_time_signature_denominator = msg_denominator
                cur_bar_capacity_total = int(
                    self.ppqn * 4 * cur_time_signature_numerator / cur_time_signature_denominator)
                cur_bar_capacity_remaining = cur_bar_capacity_total

                tokens.append(
                    f"{TokenisationPrefixes.TIME_SIGNATURE.value}_{scaled:02}_{DEFAULT_TIME_SIGNATURE_NUMERATOR:02}")

        # Close bar and handle rest buffer
        if cur_time_bar > 0 and cur_bar_capacity_remaining > 0:
            _apply_rest(cur_bar_capacity_remaining)

        # Update state dictionary
        state_dict["cur_time"] = cur_time
        state_dict["cur_time_bar"] = cur_time_bar
        state_dict["cur_time_signature_numerator"] = cur_time_signature_numerator
        state_dict["cur_time_signature_denominator"] = cur_time_signature_denominator
        state_dict["cur_bar_capacity_remaining"] = cur_bar_capacity_remaining
        state_dict["prv_track"] = prv_track
        state_dict["prv_value"] = prv_value
        state_dict["prv_velocity"] = prv_velocity

        return tokens

    def detokenise(self,
                   tokens: list[str]) -> list[Sequence]:
        # Setup Values
        sequences = [Sequence() for _ in range(self.num_tracks)]
        cur_time = 0
        cur_time_bar = 0
        cur_time_signature_numerator = DEFAULT_TIME_SIGNATURE_NUMERATOR
        cur_time_signature_denominator = DEFAULT_TIME_SIGNATURE_DENOMINATOR
        cur_bar_capacity_total = int(self.ppqn * 4 * cur_time_signature_numerator / cur_time_signature_denominator)
        cur_bar_capacity_remaining = cur_bar_capacity_total
        prv_track = 0
        prv_value = 24
        prv_velocity = 127

        for token in tokens:
            token_parts = sorted(self._split_token(token),
                                 key=lambda part: (
                                     self.sort_order.index(part[0]) if part[0] in self.sort_order else -1))
            main_parts = [part[0] for part in token_parts]

            for i, main_part in enumerate(main_parts):
                if main_part == TokenisationPrefixes.PAD.value:
                    continue
                elif main_part == TokenisationPrefixes.START.value:
                    continue
                elif main_part == TokenisationPrefixes.STOP.value:
                    continue
                elif main_part == TokenisationPrefixes.BAR.value:
                    cur_time += cur_bar_capacity_remaining
                    cur_time_bar = 0
                    cur_bar_capacity_remaining = cur_bar_capacity_total

                    for sequence in sequences:
                        sequence.add_absolute_message(Message(message_type=MessageType.INTERNAL, time=cur_time))
                elif main_part == TokenisationPrefixes.REST.value:
                    cur_time += int(token_parts[i][1])
                    cur_time_bar += int(token_parts[i][1])
                    cur_bar_capacity_remaining -= int(token_parts[i][1])
                elif main_part == TokenisationPrefixes.POSITION.value:
                    nxt_time_bar = int(token_parts[i][1])
                    nxt_time_shift = nxt_time_bar - cur_time_bar

                    if nxt_time_bar < cur_time_bar:
                        LOGGER.info("Skipping backward bar position")
                    else:
                        cur_time += nxt_time_shift
                        cur_time_bar += nxt_time_shift
                        cur_bar_capacity_remaining -= nxt_time_shift
                elif main_part == TokenisationPrefixes.TRACK.value:
                    prv_track = int(token_parts[i][1])
                elif main_part == TokenisationPrefixes.VALUE.value:
                    prv_value = int(token_parts[i][1])
                elif main_part == TokenisationPrefixes.VELOCITY.value:
                    prv_velocity = int(token_parts[i][1])
                elif main_part == TokenisationPrefixes.PITCH.value:
                    note_pitch = int(token_parts[i][1])

                    sequences[prv_track].add_absolute_message(
                        Message(message_type=MessageType.NOTE_ON, note=note_pitch, time=cur_time,
                                velocity=prv_velocity)
                    )
                    sequences[prv_track].add_absolute_message(
                        Message(message_type=MessageType.NOTE_OFF, note=note_pitch, time=cur_time + prv_value)
                    )
                elif main_part == TokenisationPrefixes.TIME_SIGNATURE.value:
                    if cur_time_bar > 0:
                        LOGGER.info(
                            f"Skipping time signature change mid-bar at time {cur_time} (bar time {cur_time_bar})")
                    else:
                        switched = False

                        new_time_signature_numerator = int(token_parts[i][1])
                        new_time_signature_denominator = int(token_parts[i][2])

                        if cur_time_signature_numerator != new_time_signature_numerator or \
                                cur_time_signature_denominator != new_time_signature_denominator:
                            switched = True

                        cur_time_signature_numerator = new_time_signature_numerator
                        cur_time_signature_denominator = new_time_signature_denominator
                        cur_bar_capacity_total = int(
                            self.ppqn * 4 * cur_time_signature_numerator / cur_time_signature_denominator)
                        cur_bar_capacity_remaining = cur_bar_capacity_total

                        if self.flag_simplify_time_signature and \
                                cur_time_signature_numerator % 2 == 0 and \
                                cur_time_signature_denominator % 2 == 0:
                            cur_time_signature_numerator = int(cur_time_signature_numerator / 2)
                            cur_time_signature_denominator = int(cur_time_signature_denominator / 2)

                        if switched or not self.flag_running_values:
                            sequences[0].add_absolute_message(
                                Message(message_type=MessageType.TIME_SIGNATURE,
                                        time=cur_time,
                                        numerator=cur_time_signature_numerator,
                                        denominator=cur_time_signature_denominator)
                            )
                else:
                    raise TokenisationException(f"Invalid token: {token}")

        return sequences

    def encode(self,
               tokens: list[str]) -> list[int]:
        return [self.dictionary[token] for token in tokens]

    def decode(self,
               tokens: list[int]) -> list[str]:
        return [self.inverse_dictionary[token] for token in tokens]

    def get_info(self,
                 tokens: list[str],
                 flag_impute_values: bool = False) -> dict[str, list[int]]:
        info_pos = []
        info_pos_bar = []
        info_time = []
        info_time_bar = []
        info_pitch = []
        info_cof = []

        # Setup Values
        cur_pos = 0
        cur_pos_bar = 0
        cur_time = 0
        cur_time_bar = 0
        cur_time_signature_numerator = DEFAULT_TIME_SIGNATURE_NUMERATOR
        cur_time_signature_denominator = DEFAULT_TIME_SIGNATURE_DENOMINATOR
        cur_bar_capacity_total = int(self.ppqn * 4 * cur_time_signature_numerator / cur_time_signature_denominator)
        cur_bar_capacity_remaining = cur_bar_capacity_total
        prv_pitch = 69  # Concert pitch A4

        for token in tokens:
            token_parts = sorted(self._split_token(token),
                                 key=lambda part: (
                                     self.sort_order.index(part[0]) if part[0] in self.sort_order else -1))
            main_parts = [part[0] for part in token_parts]
            main_part = main_parts[0]

            info_pos.append(cur_pos)
            info_pos_bar.append(cur_pos_bar)
            info_time.append(cur_time)
            info_time_bar.append(cur_time_bar)

            cur_pos += 1
            cur_pos_bar += 1

            if main_part == TokenisationPrefixes.BAR.value:
                cur_time += cur_bar_capacity_remaining
                cur_pos_bar = 0
                cur_time_bar = 0
                cur_bar_capacity_remaining = cur_bar_capacity_total

                if not flag_impute_values:
                    info_pitch.append(math.nan)
                    info_cof.append(math.nan)
                else:
                    info_pitch.append(prv_pitch)
                    info_cof.append(CircleOfFifths.get_position(prv_pitch))
            elif main_part == TokenisationPrefixes.REST.value:
                cur_time += int(token_parts[0][1])
                cur_time_bar += int(token_parts[0][1])
                cur_bar_capacity_remaining -= int(token_parts[0][1])

                if not flag_impute_values:
                    info_pitch.append(math.nan)
                    info_cof.append(math.nan)
                else:
                    info_pitch.append(prv_pitch)
                    info_cof.append(CircleOfFifths.get_position(prv_pitch))
            elif main_part == TokenisationPrefixes.POSITION.value:
                nxt_time_bar = int(token_parts[0][1])
                nxt_time_shift = nxt_time_bar - cur_time_bar
                cur_time += nxt_time_shift
                cur_time_bar += nxt_time_shift
                cur_bar_capacity_remaining -= nxt_time_shift

                if not flag_impute_values:
                    info_pitch.append(math.nan)
                    info_cof.append(math.nan)
                else:
                    info_pitch.append(prv_pitch)
                    info_cof.append(CircleOfFifths.get_position(prv_pitch))
            elif TokenisationPrefixes.PITCH.value in main_parts:
                pitch_part = next(part for part in token_parts if part[0] == TokenisationPrefixes.PITCH.value)
                note_pitch = int(pitch_part[1])

                prv_pitch = note_pitch
                info_pitch.append(note_pitch)
                info_cof.append(CircleOfFifths.get_position(note_pitch))
            elif main_part == TokenisationPrefixes.TIME_SIGNATURE.value:
                if cur_time_bar > 0:
                    LOGGER.info(
                        f"Skipping time signature change mid-bar at time {cur_time} (bar time {cur_time_bar})")
                else:
                    cur_time_signature_numerator = int(token_parts[0][1])
                    cur_time_signature_denominator = int(token_parts[0][2])
                    cur_bar_capacity_total = int(
                        self.ppqn * 4 * cur_time_signature_numerator / cur_time_signature_denominator)
                    cur_bar_capacity_remaining = cur_bar_capacity_total

                if not flag_impute_values:
                    info_pitch.append(math.nan)
                    info_cof.append(math.nan)
                else:
                    info_pitch.append(prv_pitch)
                    info_cof.append(CircleOfFifths.get_position(prv_pitch))
            else:
                if not flag_impute_values:
                    info_pitch.append(math.nan)
                    info_cof.append(math.nan)
                else:
                    info_pitch.append(prv_pitch)
                    info_cof.append(CircleOfFifths.get_position(prv_pitch))

        return {"info_position": info_pos,
                "info_position_bar": info_pos_bar,
                "info_time": info_time,
                "info_time_bar": info_time_bar,
                "info_pitch": info_pitch,
                "info_circle_of_fifths": info_cof}

    def get_hierarchy_transitions(
            self,
            tokens: list[str],
            intervals_ticks: list[int],
            seq_len: int = None,
            padding_value: int = -1
    ) -> list[list[int]]:
        """Compute hierarchy segment assignments from tick-based intervals.

        Maps each token position to a segment ID at each hierarchy level based on
        the token's absolute tick time. Segment ID is computed as tick_time // interval.

        This method is designed for use with hierarchical transformer models like HierarTune,
        which use segment assignments to determine where hierarchy tokens should be inserted
        and how attention should be structured across hierarchy levels.

        Args:
            tokens: List of string tokens (output from tokenise())
            intervals_ticks: List of tick intervals for each hierarchy level,
                            ordered finest to coarsest. E.g., [24, 96, 384] for
                            beats/bars/4-bar-blocks at PPQN=24
            seq_len: Target sequence length for padding. If None, uses len(tokens).
                    Positions beyond len(tokens) will be filled with padding_value.
            padding_value: Value for padding positions (default: -1). The value -1 is
                          commonly used by hierarchical models to indicate invalid positions.

        Returns:
            List[List[int]]: assignments[level][position] = segment_id

            Format matches HierarTune's expected tensor shape when converted:
            torch.tensor(result) -> shape (num_levels, seq_len)

            Transitions between segments occur when consecutive valid positions have
            different segment IDs.

        Example:
            >>> tokeniser = MultiTrackLargeVocabularyNotelikeTokeniser(ppqn=24)
            >>> # ... tokenise a sequence ...
            >>> intervals = [24, 96, 384]  # beats, bars, 4-bar phrases
            >>> assignments = tokeniser.get_hierarchy_transitions(tokens, intervals)
            >>> # assignments[0] = beat segments:   [0, 0, 1, 1, 2, 2, ...]
            >>> # assignments[1] = bar segments:    [0, 0, 0, 0, 1, 1, ...]
            >>> # assignments[2] = phrase segments: [0, 0, 0, 0, 0, 0, ...]

        """
        # Get timing information for each token
        info = self.get_info(tokens)
        info_time = info["info_time"]

        # Determine lengths
        actual_len = len(tokens)
        target_len = seq_len if seq_len is not None else actual_len

        # Build assignments for each level
        hierarchy_assignments = []
        for interval in intervals_ticks:
            level_assignments = []
            for pos in range(target_len):
                if pos < actual_len:
                    tick_time = info_time[pos]
                    segment_id = tick_time // interval
                    level_assignments.append(int(segment_id))
                else:
                    level_assignments.append(padding_value)
            hierarchy_assignments.append(level_assignments)

        return hierarchy_assignments

    @staticmethod
    def _split_token(token: str):
        parts = token.split("-")
        sub_parts = [part.split("_") for part in parts]
        return sub_parts

    def _construct_dictionary(self):
        self.dictionary[TokenisationPrefixes.PAD.value] = self.dictionary_size
        self._dictionary_size += 1

        self.dictionary[TokenisationPrefixes.START.value] = self.dictionary_size
        self._dictionary_size += 1

        self.dictionary[TokenisationPrefixes.STOP.value] = self.dictionary_size
        self._dictionary_size += 1

        if self.flag_bar_token:
            self.dictionary[TokenisationPrefixes.BAR.value] = self.dictionary_size
            self._dictionary_size += 1

        if not self.flag_absolute_bar_position:
            for step_size in self.step_sizes:
                self.dictionary[f"{TokenisationPrefixes.REST.value}_{step_size:02}"] = self.dictionary_size
                self._dictionary_size += 1
        else:
            for position in range(1, int(self.bar_position_quarters_range * self.ppqn)):
                self.dictionary[f"{TokenisationPrefixes.POSITION.value}_{position:03}"] = self.dictionary_size
                self._dictionary_size += 1

        combinations = []

        if self.flag_fuse_track:
            combinations.append([track for track in range(self.num_tracks)])
        else:
            for track in range(self.num_tracks):
                self.dictionary[f"{TokenisationPrefixes.TRACK.value}_{track:02}"] = self.dictionary_size
                self._dictionary_size += 1
        combinations.append([pitch for pitch in range(self.pitch_range[0], self.pitch_range[1] + 1)])
        if self.flag_fuse_value:
            combinations.append([note_value for note_value in self.note_values])
        else:
            for note_value in self.note_values:
                self.dictionary[f"{TokenisationPrefixes.VALUE.value}_{note_value:02}"] = self.dictionary_size
                self._dictionary_size += 1
        if self.flag_fuse_velocity:
            combinations.append([velocity_bin for velocity_bin in self.velocity_bins])
        else:
            for velocity_bin in self.velocity_bins:
                self.dictionary[f"{TokenisationPrefixes.VELOCITY.value}_{velocity_bin:03}"] = self.dictionary_size
                self._dictionary_size += 1

        for combination in itertools.product(*combinations):
            parts = list(combination)
            token = ""

            if self.flag_fuse_track:
                token += f"{TokenisationPrefixes.TRACK.value}_{parts.pop(0):02}-"

            token += f"{TokenisationPrefixes.PITCH.value}_{parts.pop(0):03}-"

            if self.flag_fuse_value:
                token += f"{TokenisationPrefixes.VALUE.value}_{parts.pop(0):02}-"

            if self.flag_fuse_velocity:
                token += f"{TokenisationPrefixes.VELOCITY.value}_{parts.pop(0):03}"

            self.dictionary[token] = self.dictionary_size
            self._dictionary_size += 1

        if self.time_signature_range is not None and len(self.time_signature_range) > 0:
            for time_signature in range(self.time_signature_range[0], self.time_signature_range[1] + 1):
                self.dictionary[
                    f"{TokenisationPrefixes.TIME_SIGNATURE.value}_{time_signature:02}_{DEFAULT_TIME_SIGNATURE_DENOMINATOR:02}"] = self.dictionary_size
                self._dictionary_size += 1

        self.inverse_dictionary = {v: k for k, v in self.dictionary.items()}



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.exceptions.sequence_exception import SequenceException
from scoda.misc.scoda_logging import get_logger
from scoda.misc.util import binary_insort, find_minimal_distance, get_default_step_sizes, get_default_note_values
from scoda.sequences.abstract_sequence import AbstractSequence
from scoda.settings.settings import PPQN

if TYPE_CHECKING:
    from scoda.sequences.relative_sequence import RelativeSequence


class AbsoluteSequence(AbstractSequence):
    """Class representing a sequence with absolute message timings.
    """

    LOGGER = get_logger(__name__)

    # General Methods

    def __init__(self, messages: list = None) -> None:
        super().__init__(messages=messages)

    def __eq__(self, o: object) -> bool:
        return self.equals(o)

    def to_relative_sequence(self) -> RelativeSequence:
        """Converts this AbsoluteSequence to a RelativeSequence.

        Returns: The relative representation of this sequence.

        """
        from scoda.sequences.relative_sequence import RelativeSequence
        relative_sequence = RelativeSequence()
        current_point_in_time = 0

        for msg in self._messages:
            time = msg.time
            # Check if we have to add wait messages
            if time > current_point_in_time:
                relative_sequence.add_message(
                    Message(message_type=MessageType.WAIT, channel=msg.channel, time=time - current_point_in_time))
                current_point_in_time = time

            if msg.message_type != MessageType.INTERNAL:
                message_to_add = msg.copy()
                message_to_add.time = None
                relative_sequence.add_message(message_to_add)

        return relative_sequence

    # Basic Methods

    def add_message(self, msg: Message) -> None:
        """Adds the given message to the current sequence."""
        binary_insort(self._messages, msg)

    def _add_message_unsorted(self, msg: Message) -> None:
        """Adds the given message to the current sequence."""
        self._messages.append(msg)

    def cutoff(self, maximum_length, reduced_length) -> None:
        """Reduces the length of all notes longer than the maximum length to this value.

        Args:
            maximum_length: Maximum note length allowed in this sequence.
            reduced_length: The length violating notes are assigned.

        """
        channel_pairings = self.get_message_pairings()

        for message_pairings in channel_pairings.values():
            for message_pairing in message_pairings:
                if len(message_pairing) == 1:
                    if not message_pairing[0].message_type == MessageType.NOTE_ON:
                        raise SequenceException("Cutoff: Note was closed without having been opened.")
                    self.add_message(
                        Message(message_type=MessageType.NOTE_OFF, channel=message_pairing[0].channel,
                                note=message_pairing[0].note,
                                time=message_pairing[0].time + maximum_length))
                else:
                    if message_pairing[1].time - message_pairing[0].time > maximum_length:
                        message_pairing[1].time = message_pairing[0].time + reduced_length

        self.normalise_absolute()

    def equals(self,
               other: object,
               ignore_channel: bool = False,
               ignore_time_signature: bool = False,
               ignore_key_signature: bool = False,
               ignore_velocity: bool = False) -> bool:
        """Checks if this sequence is equal to another one.

        Args:
            other: The other sequence to compare with.
            ignore_channel: If the channel should be ignored.
            ignore_time_signature: If the time signature should be ignored.
            ignore_key_signature: If the key signature should be ignored.
            ignore_velocity: If the velocity should be ignored

        Returns: True if the sequences are equal, False otherwise.

        """
        if not isinstance(other, AbsoluteSequence):
            return False

        # Construct pairings
        message_types = [MessageType.NOTE_ON, MessageType.NOTE_OFF,
                         MessageType.TIME_SIGNATURE, MessageType.KEY_SIGNATURE]

        if ignore_time_signature:
            message_types.remove(MessageType.TIME_SIGNATURE)
        if ignore_key_signature:
            message_types.remove(MessageType.KEY_SIGNATURE)

        self_pairings = self.get_interleaved_message_pairings(message_types=message_types)
        other_pairings = other.get_interleaved_message_pairings(message_types=message_types)

        if not len(self_pairings) == len(other_pairings):
            return False

        for self_pair, other_pair in zip(self_pairings, other_pairings):
            self_channel = self_pair[0]
            other_channel = other_pair[0]

            # Compare channels
            if self_channel != other_channel and not ignore_channel:
                return False

            self_msgs = self_pair[1]
            other_msgs = other_pair[1]

            self_msg = self_msgs[0]
            other_msg = other_msgs[0]

            if self_msg.message_type != other_msg.message_type:
                return False

            if self_msg.message_type == MessageType.NOTE_ON:
                self_msg_value = self_msgs[1].time - self_msg.time
                other_msg_value = other_msgs[1].time - other_msg.time

                if self_msg.note != other_msg.note or self_msg_value != other_msg_value:
                    return False
                if self_msg.velocity != other_msg.velocity and not ignore_velocity:
                    return False
            elif self_msg.message_type == MessageType.TIME_SIGNATURE:
                if self_msg.numerator != other_msg.numerator or self_msg.denominator != other_msg.denominator:
                    return False
            elif self_msg.message_type == MessageType.KEY_SIGNATURE:
                if self_msg.key != other_msg.key:
                    return False

        return True

    def merge(self, sequences: list[AbsoluteSequence]) -> None:
        """Merges this sequence with all the given ones.

        In case of the creation of overlapping notes, these will be combined. The earliest start and the latest end will
        be used for the newly created note.

        Args:
            sequences: The sequence to merge with this one.

        """
        for sequence in sequences:
            for msg in [msg for msg in sequence._messages]:
                self._add_message_unsorted(msg)

        self.normalise_absolute()

    def normalise_absolute(self) -> None:
        self.sort()

    def quantise(self, step_sizes: list[int] = None) -> None:
        """Quantises the sequence to a given grid.

        Quantises the sequence stored in this object according to the given step sizes. These step sizes determine the
        size of the underlying grid, e.g. a step size of 3 would allow for messages to be placed at multiples of 3
        ticks. Note that the induced length of the notes is dependent on the `PPQN`, e.g., with a `PPQN` of 24,
        a step size of 3 would be equivalent to a grid conforming to thirty-second notes. If there exists a tie between
        two grid boundaries, these are first resolved by whether the quantisation would prevent a note-length of 0,
        then by the order of the `step_sizes` array. The result of this operation is that all messages of this
        sequence have a time divisible by one of the values in `step_sizes`. If the quantisation resulted in two
        notes overlapping, the second note will be removed. See `scoda.utils.utils.get_note_durations`,
        `scoda.utils.utils.get_tuplet_durations` and `scoda.utils.utils.get_dotted_note_durations` for generating the
        `step_sizes` array.

        Args:
            step_sizes: Array of numbers corresponding to divisors of the grid length.

        """
        if step_sizes is None:
            step_sizes = get_default_step_sizes()

        # List of finally quantised messages
        quantised_messages = []
        # Keep track of open messages, in order to guarantee quantisation does not smother them
        open_messages = dict()
        # Keep track of from when to when notes are played, in order to eliminate double notes
        message_timings = dict()

        for msg in self._messages:
            message_original_time = msg.time
            message_to_append = msg

            # Positions the note would land at according to each of the quantisation parameters
            positions_left = [(message_original_time // step_size) * step_size for step_size in step_sizes]
            positions_right = [positions_left[i] + step_sizes[i] for i in range(0, len(step_sizes))]

            possible_positions = positions_left + positions_right
            valid_positions = []

            # Consider quantisations that could smother notes
            if msg.message_type == MessageType.NOTE_ON:
                valid_positions += possible_positions
                message_to_append.time = valid_positions[find_minimal_distance(message_original_time, valid_positions)]

                # Check if note was not yet closed
                if msg.note in open_messages:
                    AbsoluteSequence.LOGGER.info(f"Quantisation: Note {msg.note} not previously stopped.")
                    quantised_messages.append(
                        Message(message_type=MessageType.NOTE_OFF, channel=msg.channel, note=msg.note,
                                time=message_to_append.time))
                    open_messages.pop(msg.note, None)
                    message_timings[msg.note].append(message_to_append.time)

                # Check if we can open note without overlaps
                if msg.note not in message_timings \
                        or not message_to_append.time < message_timings[msg.note][1]:
                    open_messages[msg.note] = message_to_append.time
                    message_timings[msg.note] = [message_to_append.time]
                # In this case note would overlap with other, existing note
                else:
                    message_to_append = None
            elif msg.message_type == MessageType.NOTE_OFF:
                # Message is currently open, have to quantize
                if msg.note in open_messages:
                    note_open_timing = open_messages.pop(msg.note)

                    # Add possible positions for stop messages, making sure the belonging note is not smothered
                    for position in possible_positions:
                        if not position - note_open_timing <= 0:
                            valid_positions.append(position)

                    # If no valid positions exists, set note length to 0
                    if len(valid_positions) == 0:
                        valid_positions.append(note_open_timing)

                    # Valid positions will always exist, since if order held before quantisation, same will hold
                    # after, and if initially no valid position was found note length will be set to 0
                    message_to_append.time = valid_positions[
                        find_minimal_distance(message_original_time, valid_positions)]
                    message_timings[msg.note].append(message_to_append.time)

                # Message is not currently open (e.g., if start message was removed due to an overlap)
                else:
                    message_to_append = None
            else:
                valid_positions += possible_positions
                message_to_append.time = valid_positions[find_minimal_distance(message_original_time, valid_positions)]

            if message_to_append is not None:
                quantised_messages.append(message_to_append)

        # Remove smothered notes
        message_timings_with_indices = dict()
        original_indices_to_remove = []

        # Get indices of violating messages
        for i, msg in enumerate(quantised_messages):
            if msg.message_type == MessageType.NOTE_ON:
                message_timings_with_indices[msg.note] = (i, msg.time)
            elif msg.message_type == MessageType.NOTE_OFF:
                j, time = message_timings_with_indices.pop(msg.note)
                if msg.time - time <= 0:
                    original_indices_to_remove.extend([j, i])

        # Remove messages
        for index_shifter, index_to_remove in enumerate(original_indices_to_remove):
            quantised_messages.pop(index_to_remove - index_shifter)

        self._messages = quantised_messages
        self.normalise_absolute()

    def quantise_note_lengths(self, note_values=None, standard_length=PPQN, do_not_extend=False) -> None:
        """Quantises the note lengths of this sequence, only affecting the ending of the notes.

        Quantises notes to the given values, ensuring that all notes are of one of the sizes defined by the
        parameters. See `scoda.utils.utils.get_note_durations`, `scoda.utils.utils.get_tuplet_durations` and
        `scoda.utils.utils.get_dotted_note_durations` for generating the `note_values` array. Tries to shorten
        or extend the end of each note in such a way that the note duration is exactly one of the values given in
        `note_values`. If this is not possible (e.g., due to an overlap with another note that would occur),
        the note that can neither be shortened nor lengthened will be removed from the sequence. Note that this is
        only the case if the note was shorter than the smallest legal duration specified, and thus cannot be shortened.

        Args:
            note_values: An array containing exactly the valid note durations in ticks.
            standard_length: Note length for notes which are not closed.
            do_not_extend: Determines if notes are only allowed to be shortened, e.g., for bars.

        """
        # Construct possible durations
        if note_values is None:
            note_values = get_default_note_values()

        # Construct current durations
        channel_pairings = self.get_message_pairings(standard_length=standard_length)
        quantised_messages = []

        for message_pairings in channel_pairings.values():
            # Track when each type of note occurs, in order to check for possible overlaps
            note_occurrences = dict()

            # Construct array keeping track of when each note occurs
            for message_pairing in message_pairings:
                note = message_pairing[0].note
                note_occurrences.setdefault(note, [])
                note_occurrences[note].append(message_pairing)

            # Handle each note, pairing consists of start and stop message
            for i, message_pairing in enumerate(message_pairings):
                note = message_pairing[0].note
                current_duration = message_pairing[1].time - message_pairing[0].time
                valid_durations = copy.copy(note_values)

                # Check if the current note is not the last note, in this case clashes with a next note could exist
                index = note_occurrences[note].index(message_pairing)
                if index != len(note_occurrences[note]) - 1:
                    possible_next_pairing = note_occurrences[note][index + 1]

                    # Note values contains the same as valid durations at the beginning
                    for note_value in note_values:
                        possible_correction = note_value - current_duration

                        # If we cannot extend the note, remove the time from possible times
                        if message_pairing[1].time + possible_correction > possible_next_pairing[0].time:
                            valid_durations.remove(note_value)

                # If we are not allowed to extend note lengths, remove all positive corrections
                for note_value in note_values:
                    possible_correction = note_value - current_duration

                    if possible_correction > 0 and do_not_extend and note_value in valid_durations:
                        valid_durations.remove(note_value)

                # Check if we have to remove the note
                if len(valid_durations) == 0:
                    message_pairings[i] = []
                else:
                    current_duration = message_pairing[1].time - message_pairing[0].time
                    best_fit = valid_durations[find_minimal_distance(current_duration, valid_durations)]
                    correction = best_fit - current_duration
                    message_pairing[1].time += correction

            for message_pairing in message_pairings:
                quantised_messages.extend(message_pairing)

        for msg in self._messages:
            if msg.message_type is not MessageType.NOTE_ON and msg.message_type is not MessageType.NOTE_OFF:
                quantised_messages.append(msg)

        self._messages = quantised_messages
        self.normalise_absolute()

    def similarity(self, other: AbsoluteSequence,
                   flag_consider_channel: bool = True,
                   flag_consider_program:bool = False,
                   flag_consider_velocity: bool = False) -> float:
        """Calculates a similarity measure between this sequence and another one.

        The similarity is calculated as the number of matching notes divided by the combined total number of notes in both
        sequences. Two notes are considered matching if they have the same channel, note value, start time, end time, and
        conditionally, velocity.

        Args:
            other: The other sequence to compare with.
            flag_consider_velocity: If the velocity should be considered when comparing notes.

        Returns: A value between 0 and 1, where 1 means the sequences are identical.

        """
        message_types = [MessageType.NOTE_ON, MessageType.NOTE_OFF]

        self_notes, other_notes = set(), set()
        for notes, pairings in [(self_notes, self.get_interleaved_message_pairings(message_types=message_types)),
                                (other_notes, other.get_interleaved_message_pairings(message_types=message_types))]:
            for channel, msgs in pairings:
                if len(msgs) == 2 and msgs[0].message_type == MessageType.NOTE_ON and msgs[
                    1].message_type == MessageType.NOTE_OFF:
                    note_tuple = (channel,
                                  msgs[0].note,
                                  msgs[0].time,
                                  msgs[1].time,
                                  msgs[0].channel if flag_consider_channel else None,
                                  msgs[0].program if flag_consider_program else None,
                                  msgs[0].velocity if flag_consider_velocity else None)
                    notes.add(note_tuple)

        # Check if both sets are empty
        if not self_notes and not other_notes:
            return 1.0

        matching_notes = self_notes & other_notes
        all_notes = self_notes | other_notes

        if len(all_notes) == 0:
            return 1.0

        similarity_score = len(matching_notes) / len(all_notes)
        return similarity_score

    def sort(self) -> None:
        """Sorts the sequence according to the timings of the messages.

        This sorting procedure is stable, if two messages occurred in a specific order at the same time before the sort,
        they will occur in this order after the sort.

        """
        self._messages.sort(key=lambda x: (x.time, -1 if x.channel is None else x.channel, x.message_type, x.note))

    # Misc. Methods

    def get_message_pairings(self,
                             message_types: list[MessageType] = None,
                             standard_length=PPQN,
                             impute_notes=True) -> dict[int, list[list[Message]]]:
        """Creates a dictionary where the keys correspond to channels and the items to lists of messages of the given types.
        These lists contain all entries for messages that belong together, e.g., open and close note messages.

        Args:
            message_types: All message types to include.
            standard_length: The length used for notes that have not been closed.
            impute_notes: If unclosed notes should be closed and unopened ones should be ignored.

        Returns: A dictionary of lists of messages belonging together.

        """
        if message_types is None:
            message_types = [MessageType.NOTE_ON, MessageType.NOTE_OFF]

        self.normalise_absolute()

        message_pairings = dict()
        open_messages = dict()

        # Collect notes
        for msg in self._messages:
            if msg.message_type in message_types:
                # Set defaults
                message_pairings.setdefault(msg.channel, [])
                open_messages.setdefault(msg.channel, dict())

                # Add notes to open messages
                if msg.message_type == MessageType.NOTE_ON:
                    # Check if note was not previously closed
                    if msg.note in open_messages[msg.channel] and impute_notes:
                        AbsoluteSequence.LOGGER.info(
                            f"Time Pairings: Note {msg.channel}:{msg.note} at time {msg.time} not previously stopped.")
                        index = open_messages[msg.channel].pop(msg.note)
                        message_pairings[msg.channel][index].append(
                            Message(message_type=MessageType.NOTE_OFF, channel=msg.channel, note=msg.note,
                                    time=msg.time))

                    message_pairings[msg.channel].append([msg])
                    open_messages[msg.channel][msg.note] = len(message_pairings[msg.channel]) - 1

                # Add closing message to fitting open message
                elif msg.message_type == MessageType.NOTE_OFF:
                    # Check if note was not previously started
                    if msg.channel not in open_messages or msg.note not in open_messages[msg.channel]:
                        if impute_notes:
                            AbsoluteSequence.LOGGER.info(
                                f"Time Pairings: Note {msg.channel}:{msg.note} at time {msg.time} not previously started.")
                    else:
                        index = open_messages[msg.channel].pop(msg.note)
                        message_pairings[msg.channel][index].append(msg)

                else:
                    message_pairings[msg.channel].append([msg])

        # Check unclosed notes
        for channel in message_pairings:
            for pairing in message_pairings[channel]:
                if len(pairing) == 1 and pairing[0].message_type == MessageType.NOTE_ON and impute_notes:
                    pairing.append(Message(message_type=MessageType.NOTE_OFF, channel=pairing[0].channel,
                                           note=pairing[0].note, time=pairing[0].time + standard_length))

        return message_pairings

    def get_interleaved_message_pairings(self,
                                         message_types: list[MessageType] = None,
                                         standard_length=PPQN,
                                         impute_notes=True) -> list[tuple[int, list[Message]]]:
        """Creates a list of tuples of channels and messages sorted by their absolute starting times.

        Args:
            message_types: All message types to include.
            standard_length: The length used for notes that have not been closed.
            impute_notes: If unclosed notes should be closed and unopened ones should be ignored.

        Returns: A list of channel-message pairs sorted by their start times.

        """
        interleaved_pairings = []
        channel_pairings = self.get_message_pairings(message_types=message_types,
                                                     standard_length=standard_length,
                                                     impute_notes=impute_notes)

        channel_pairings_list = list(channel_pairings.items())
        channel_cur_index = [0 for _ in range(len(channel_pairings_list))]
        channel_max_index = [len(channel_pairings_list[i][1]) for i in range(len(channel_pairings_list))]
        channel_nxt_times = [channel_pairings_list[n][1][channel_cur_index[n]][0].time
                             if channel_cur_index[n] < channel_max_index[n] else float("inf")
                             for n in range(len(channel_pairings_list))]
        channel_ids = [channel_pairings_list[i][0] for i in range(len(channel_pairings_list))]

        has_next = len(channel_pairings_list) > 0

        while has_next:
            # Build list of next times for each channel
            track_val_times = [channel_nxt_times[i] if channel_cur_index[i] < channel_max_index[i] else float('inf')
                               for i in range(len(channel_ids))]

            # Get channel ID and index of channel in list and obtain next message pairing
            next_channel_index = track_val_times.index(min(track_val_times))
            next_channel = channel_ids[next_channel_index]
            next_message_pairing = channel_pairings_list[next_channel_index][1][channel_cur_index[next_channel_index]]

            # Append message to interleaved list
            interleaved_pairings.append((next_channel, next_message_pairing))

            # Update next times and indices
            channel_cur_index[next_channel_index] += 1
            channel_nxt_times = [channel_pairings_list[n][1][channel_cur_index[n]][0].time
                                 if channel_cur_index[n] < channel_max_index[n] else float("inf")
                                 for n in range(len(channel_pairings_list))]
            has_next = any(
                channel_cur_index[i] < len(channel_pairings_list[i][1]) for i in range(len(channel_pairings_list)))

            assert next_channel == next_message_pairing[0].channel

        return interleaved_pairings

    def get_message_times_of_type(self, message_types: list[MessageType]) -> list[tuple[int, Message]]:
        """Searches for messages that fit one of the given types.

        Args:
            message_types: Which message types to search for.

        Returns: A list containing the found messages and their absolute points in time.

        """
        timings = []

        for msg in self._messages:
            if msg.message_type in message_types:
                timings.append((msg.time, msg))

        return timings

    def get_sequence_channel(self) -> int | None:
        if not self._messages:
            return None
        if not self.is_channel_consistent():
            raise SequenceException("Sequence has inconsistent channels.")

        return self._messages[0].channel

    def get_sequence_duration(self) -> int:
        """Calculates the overall duration of this sequence, given in ticks.

        Returns: The duration of this sequence.

        """
        if not self._messages:
            return 0
        return self._messages[-1].time

    def is_channel_consistent(self) -> bool:
        """Checks if the channels of all messages in this sequence are consistent, i.e., the same.

        Returns: True if the channel numbers are consistent, False otherwise.

        """
        for msg in self._messages:
            if msg.channel != self._messages[0].channel:
                return False
        return True
